
//***************************************************
// C++ document adapter for Winmeta plugins
//
// Requires: WINMETA.DLL version 1.0
// (WINMETA.DLL is  2007 Peter Koller, Maison Anglais. All Rights Reserved)
//
// Copyright (C) 2007-2008 Heiko Nitzsche
//
//   This software is provided 'as-is', without any express or implied
//   warranty.  In no event will the author be held liable for any damages
//   arising from the use of this software.
//
//   Permission is granted to anyone to use this software for any purpose,
//   including commercial applications, and to alter it and redistribute it
//   freely, subject to the following restrictions:
//
//   1. The origin of this software must not be misrepresented; you must not
//      claim that you wrote the original software. If you use this software
//      in a product, an acknowledgment in the product documentation would be
//      appreciated but is not required.
//   2. Altered source versions must be plainly marked as such, and must not be
//      misrepresented as being the original software.
//   3. This notice may not be removed or altered from any source distribution.
//
//***************************************************

#ifdef __IBMCPP__
  #pragma strings( readonly )
#endif

#include <math.h>
#include "os2defs.h"
#include "MetaDocument.hpp"

// ---------------------------------------------------------
// ---------------------------------------------------------

FIXED _doubletofixed(double in)
{
    double integer;
    LONG   ret;
    modf(0x10000 * in, &integer);
    ret = (LONG)integer;
    return((FIXED)ret);
}

// ---------------------------------------------------------
// ---------------------------------------------------------

MetaDocument::MetaDocument(const MetaAccessor & accessor)
    : mMetaAccessor(accessor),
      mMetadata(NULL),
      mMetadataSize(0),
      mhMetafile(NULLHANDLE),
      mX(0),
      mY(0),
      mWidth(0),
      mHeight(0),
      mFormatDescription(NULL)
{
}

// ---------------------------------------------------------

MetaDocument::~MetaDocument()
{
    delete [] mMetadata; mMetadata = NULL; mMetadataSize = 0;

    if (mhMetafile != NULLHANDLE)
    {
        GpiDeleteMetaFile(mhMetafile);
        mhMetafile = NULLHANDLE;
    }
}

// ---------------------------------------------------------

int MetaDocument::getWidth()
{
    return mWidth;
}

// ---------------------------------------------------------

int MetaDocument::getHeight()
{
    return mHeight;
}

// ---------------------------------------------------------

BOOL MetaDocument::hasContent() const
{
    return (mhMetafile != NULLHANDLE);
}

// ---------------------------------------------------------

void MetaDocument::setDocumentFile(const char * filename) throw ( MetaException )
{
    delete [] mMetadata; mMetadata = NULL; mMetadataSize = 0;
    mX      = 0;
    mY      = 0;
    mWidth  = 0;
    mHeight = 0;
    mFormatDescription = NULL;

    if (mhMetafile != NULLHANDLE)
    {
        GpiDeleteMetaFile(mhMetafile);
        mhMetafile = NULLHANDLE;
    }

    // read file (throws MetaException in case of an error)
    readFile(filename);

    HWND hwnd = WinQueryActiveWindow(HWND_DESKTOP);
    HAB  hab  = WinQueryAnchorBlock(hwnd ? hwnd : HWND_DESKTOP);
    HMF  hmf  = NULLHANDLE;

    // check if this is a Windows Metafile
    if (mMetaAccessor.IsWinMetafile(mMetadata, mMetadataSize))
    {
        hmf = importWindowsMetafile(hab);
        if (hmf == GPI_ERROR)
        {
           delete [] mMetadata; mMetadata = NULL; mMetadataSize = 0;
           throw MetaException(MetaException::META_ERR_NOT_SUPP,
                               "File format is not supported");
        }
        mFormatDescription = "Windows Metafile";
    }
    else
    {
        hmf = importOS2Metafile(hab);
        if (hmf == GPI_ERROR)
        {
            delete [] mMetadata; mMetadata = NULL; mMetadataSize = 0;
            throw MetaException(MetaException::META_ERR_NOT_SUPP,
                                "File format is not supported");
        }
        mFormatDescription = "OS/2 Metafile";
    }

    mhMetafile = hmf;

    // calc the original size of the metafile
    calculateMetafileSize(hab);
}

// ---------------------------------------------------------

void MetaDocument::writeDocumentToFile(const char * filename) throw ( MetaException )
{
    // write file (throws MetaException in case of an error)
    writeFile(filename);
}

// ---------------------------------------------------------

/**
 * @param src_x         x coordinate of upper left corner
 * @param src_y         y coordinate of upper left corner
 * @param src_width     width of rectangle to render
 * @param src_height    height of rectangle to render
 * @param scale_x       scale specified as pixels per point (width)
 * @param scale_y       scale specified as pixels per point (height)
 * @param hps           Presentation Space to render on
 *
 * 1. Scale the document to match the specified pixels per point
 * 2. Render the rectangle given by the upper left corner at
 *    (src_x, src_y) and src_width and src_height.
 **/

void MetaDocument::renderToHPS(long   src_x    , long   src_y,
                               long   src_width, long   src_height,
                               double scale_x  , double scale_y,
                               long   rotation ,
                               HPS    hps) throw ( MetaException )
{
    if (! hasContent())
    {
        return;
    }
    if ((scale_x < 0.0) || (scale_y < 0.0) ||
        (src_y < 0)     || (src_x < 0)     || (src_width <= 0) || (src_height <= 0))
    {
        throw MetaException(MetaException::META_ERR_BAD_ARG,
                            "Bad render options");
    }

    LONG     lpsid = GpiSavePS(hps);
    LONG     opts[PMF_DEFAULTS + 1] = {0};
    MATRIXLF xformArray, oldArray;
    FIXED    afxScale[2]      = {0,0};
    POINTL   pointScaleOffset = {0,0};
    unsigned char desc[40];

    setOpts(opts, MetaDocument::SETOPTS_PAINT);

    if (! GpiQueryDefaultViewMatrix(hps, 9, &xformArray))
    {
        throw MetaException(MetaException::META_ERR_RENDER,
                            "Querying default view matrix failed");
    }
    oldArray = xformArray;

#if 0
    // set up scaling
    const LONG scaledWidth  = (LONG)(scale_x * ((double) mWidth));
    const LONG scaledHeight = (LONG)(scale_y * ((double) mHeight));
#endif



----------------

    pointScaleOffset.x = src_x - mX;
    pointScaleOffset.y = src_y - mY;
    afxScale[0]        = MAKEFIXED((int)scale_x, 0); // TODO
    afxScale[1]        = MAKEFIXED((int)scale_y, 0); // TODO

    // setup translation
    if (! GpiTranslate(hps, &xformArray, TRANSFORM_REPLACE, &pointScaleOffset))
    {
        throw MetaException(MetaException::META_ERR_RENDER,
                            "Translation failed");
    }

    // scale
    pointScaleOffset.x = src_x;
    pointScaleOffset.y = src_y;

    if (! GpiScale(hps,
                   &xformArray,
                   TRANSFORM_ADD,
                   afxScale,
                   &pointScaleOffset))
    {
        throw MetaException(MetaException::META_ERR_RENDER,
                            "Scaling failed");
    }
/*
    // rotate
    if (! GpiRotate(hps,
                    &xformArray,
                    TRANSFORM_ADD,
                    MAKEFIXED(-rotation, 0),
                    &pointRotOffset))
    {
        throw MetaException(MetaException::META_ERR_RENDER,
                            "Rotation failed");
    }
*/
    if (! GpiSetDefaultViewMatrix(hps, 9, &xformArray, TRANSFORM_REPLACE))
    {
        GpiRestorePS(hps, lpsid);
        throw MetaException(MetaException::META_ERR_RENDER,
                            "Setting default view matrix failed");
    }

    if (GpiPlayMetaFile(hps,
                        mhMetafile,
                        PMF_DEFAULTS + 1,
                        opts,
                        NULL,
                        sizeof(desc)-1,
                        (PSZ)desc) == GPI_ERROR)
    {
        GpiSetDefaultViewMatrix(hps, 9, &oldArray, TRANSFORM_REPLACE);
        GpiRestorePS(hps, lpsid);
        throw MetaException(MetaException::META_ERR_RENDER,
                            "Setting default view matrix failed");
    }

    // reset all
    GpiSetDefaultViewMatrix(hps, 9, &oldArray, TRANSFORM_REPLACE);
    GpiRestorePS(hps, lpsid);
}

// ---------------------------------------------------------

const MetaAccessor & MetaDocument::getMetaAccessor()
{
    return mMetaAccessor;
}

// ---------------------------------------------------------

const char * MetaDocument::getDescription() const
{
    if (mFormatDescription != NULL)
    {
        return mFormatDescription;
    }
    return "";
}

// ---------------------------------------------------------
// ---------------------------------------------------------

void MetaDocument::readFile(const char * filename) throw ( MetaException )
{
   // read the file into a local buffer
   HFILE       hMetafile          = NULLHANDLE;
   FILESTATUS3 metafileInfo       = { { 0 } };
   ULONG       metafileInfoSize   = sizeof(FILESTATUS3);
   ULONG       metafileOpenAction = 0;

   APIRET rc = DosOpen((PSZ)filename,
                       &hMetafile,
                       &metafileOpenAction,
                       0L,
                       FILE_NORMAL,
                       OPEN_ACTION_OPEN_IF_EXISTS,
                       OPEN_FLAGS_SEQUENTIAL |
                       OPEN_FLAGS_NOINHERIT  |
                       OPEN_ACCESS_READONLY  |
                       OPEN_SHARE_DENYWRITE,
                       0L);
    if (rc != NO_ERROR)
    {
       throw MetaException(MetaException::META_ERR_NOT_FOUND,
                           "Cannot open file");
    }
    rc = DosQueryFileInfo(hMetafile,
                          FIL_STANDARD,
                          &metafileInfo,
                          metafileInfoSize);
    if (rc != NO_ERROR)
    {
       throw MetaException(MetaException::META_ERR_READ,
                           "Cannot read file");
    }

    // delete the old buffer, just in case multiple calls arrive
    delete [] mMetadata; mMetadata = NULL; mMetadataSize = 0;
    try
    {
        mMetadata = new char[metafileInfo.cbFile];
    }
    catch(...)
    {
    }
    // for compiler that don't throw std::bad_allow (e.g. IBM VAC)
    if (mMetadata == NULL)
    {
       DosClose(hMetafile);
       throw MetaException(MetaException::META_ERR_MEM,
                           "Out of memory");
    }

    mMetadataSize = 0L;
    rc = DosRead(hMetafile, mMetadata, metafileInfo.cbFile, &mMetadataSize);

    DosClose(hMetafile);
    hMetafile = NULLHANDLE;

    if ((rc != NO_ERROR) || (mMetadataSize != metafileInfo.cbFile))
    {
       delete [] mMetadata; mMetadata = NULL; mMetadataSize = 0;
       throw MetaException(MetaException::META_ERR_READ,
                           "Cannot read file");
    }
}

// ---------------------------------------------------------

void MetaDocument::writeFile(const char * filename) const
                             throw ( MetaException )
{
   // check that there are any data loaded
   if (mMetadata == NULL)
   {
      throw MetaException(MetaException::META_ERR_WRITE,
                          "No document data");
   }

   // write the local buffer into a file
   HFILE hMetafile          = NULLHANDLE;
   ULONG metafileOpenAction = 0;

   APIRET rc = DosOpen((PSZ)filename,
                       &hMetafile,
                       &metafileOpenAction,
                       0,
                       FILE_NORMAL,
                       OPEN_ACTION_CREATE_IF_NEW |
                       OPEN_ACTION_REPLACE_IF_EXISTS,
                       OPEN_FLAGS_SEQUENTIAL |
                       OPEN_FLAGS_NOINHERIT  |
                       OPEN_ACCESS_WRITEONLY |
                       OPEN_SHARE_DENYWRITE,
                       0L);
    if (rc != NO_ERROR)
    {
       throw MetaException(MetaException::META_ERR_WRITE,
                           "Cannot create file");
    }

    ULONG bytesWritten = 0L;
    rc = DosWrite(hMetafile, mMetadata, mMetadataSize, &bytesWritten);

    DosClose(hMetafile);
    hMetafile = NULLHANDLE;

    if ((rc != NO_ERROR) || (bytesWritten != mMetadataSize))
    {
       throw MetaException(MetaException::META_ERR_WRITE,
                           "Cannot write file");
    }
}

// ---------------------------------------------------------

HMF MetaDocument::importOS2Metafile(HAB hab)
{
    DEVOPENSTRUC dop   = {0};
//    SIZEL        sizel = {0,0};

    dop.pszDriverName = (PSZ)"DISPLAY";
    HDC metaHdc = DevOpenDC(hab, OD_METAFILE, (PSZ)"*", 5, (PDEVOPENDATA)&dop, 0L);
    if (!metaHdc || (metaHdc == DEV_ERROR))
    {
        return GPI_ERROR;
    }

    HMF hmf = DevCloseDC(metaHdc);

    if ((hmf == DEV_OK) || (hmf == DEV_ERROR))
    {
        return GPI_ERROR;
    }
    if (! GpiSetMetaFileBits(hmf, 0, mMetadataSize, (PBYTE)mMetadata))
    {
        GpiDeleteMetaFile(hmf);
        return GPI_ERROR;
    }

    return hmf;
}

// ---------------------------------------------------------

HMF MetaDocument::importWindowsMetafile(HAB hab)
{
    DEVOPENSTRUC dop   = {0};
    SIZEL        sizel = {0,0};

    dop.pszDriverName = (PSZ)"DISPLAY";
    HDC metaHdc = DevOpenDC(hab, OD_METAFILE, (PSZ)"*", 5, (PDEVOPENDATA)&dop, 0L);
    if (!metaHdc || (metaHdc == DEV_ERROR))
    {
        return GPI_ERROR;
    }

    HPS metaHps = GpiCreatePS(hab, metaHdc, &sizel,
                             PU_PELS | GPIF_DEFAULT | GPIT_NORMAL | GPIA_ASSOC);
    if (metaHps == GPI_ERROR)
    {
        DevCloseDC(metaHdc);
        return GPI_ERROR;
    }

    ULONG errcode = mMetaAccessor.InitWinMetafile(hab,
                                                  metaHps,
                                                  NULL,
                                                  0,
                                                  WINGDI_DEFAULTOPTS);
    if (errcode)
    {
        GpiAssociate(metaHps, NULLHANDLE);
        GpiDestroyPS(metaHps);
        DevCloseDC(metaHdc);
        return GPI_ERROR;
    }
    errcode = mMetaAccessor.PaintWinMetafile(metaHps, mMetadata, mMetadataSize);
    if (errcode)
    {
        GpiAssociate(metaHps, NULLHANDLE);
        GpiDestroyPS(metaHps);
        DevCloseDC(metaHdc);
        return GPI_ERROR;
    }

    GpiAssociate(metaHps, NULLHANDLE);
    GpiDestroyPS(metaHps);

    HMF hmf = DevCloseDC(metaHdc);

    return hmf;
}

// ---------------------------------------------------------

void MetaDocument::calculateMetafileSize(HAB hab)
{
    DEVOPENSTRUC  dop = {0};
    LONG          lhits;
    LONG          opts[PMF_DEFAULTS + 1] = {0};
    RECTL         rect     = {0};
    SIZEL         sizel    = {0,0};
    unsigned char desc[40] = {0};

    mX      = 0;
    mY      = 0;
    mWidth  = 0;
    mHeight = 0;

    if (mhMetafile == NULLHANDLE)
    {
        return;
    }

    dop.pszDriverName = (PSZ)"DISPLAY";
    HDC hdc = DevOpenDC(hab, OD_DIRECT, (PSZ)"*", 9, (PDEVOPENDATA)&dop, 0L);
    if (!hdc || (hdc == DEV_ERROR))
    {
        return;
    }

    HPS hps = GpiCreatePS(hab, hdc, &sizel,
                          PU_PELS | GPIF_DEFAULT | GPIT_NORMAL | GPIA_ASSOC);
    if (hps == GPI_ERROR)
    {
        DevCloseDC(hdc);
        return;
    }

    GpiSavePS(hps);

    setOpts(opts, MetaDocument::SETOPTS_GETBOUNDS);

    lhits = GpiPlayMetaFile(hps,
                            mhMetafile,
                            PMF_DEFAULTS + 1,
                            opts,
                            NULL,
                            sizeof(desc)-1,
                            (PSZ)desc);
    if (lhits == GPI_ERROR)
    {
        GpiAssociate(hps, NULLHANDLE);
        GpiDestroyPS(hps);
        DevCloseDC  (hdc);
        return;
    }

    GpiQueryPS  (hps, &sizel);
    GpiRestorePS(hps, -1);

    GpiSetDrawControl(hps, DCTL_DISPLAY , DCTL_OFF);
    GpiSetDrawControl(hps, DCTL_BOUNDARY, DCTL_ON);

    GpiResetBoundaryData(hps);

    setOpts(opts, MetaDocument::SETOPTS_GETSIZE);

    lhits = GpiPlayMetaFile(hps,
                            mhMetafile,
                            PMF_DEFAULTS + 1,
                            opts,
                            NULL,
                            sizeof(desc)-1,
                            (PSZ)desc);
    if (lhits == GPI_ERROR)
    {
        GpiAssociate(hps, NULLHANDLE);
        GpiDestroyPS(hps);
        DevCloseDC  (hdc);
        return;
    }
    GpiQueryBoundaryData(hps, &rect);
    GpiConvert(hps, CVTC_MODEL, CVTC_DEVICE, 2L, (PPOINTL)&rect);

    GpiSetDrawControl(hps, DCTL_DISPLAY, DCTL_ON);
    GpiSetDrawControl(hps, DCTL_BOUNDARY, DCTL_OFF);

    //mX      = 0;
    //mY      = 0;
    //mWidth  = sizel.cx;
    //mHeight = sizel.cy;

    mX      = rect.xLeft;
    mY      = rect.yBottom;
    mWidth  = rect.xRight - rect.xLeft;
    mHeight = rect.yTop - rect.yBottom;

    GpiAssociate(hps, NULLHANDLE);
    GpiDestroyPS(hps);
    DevCloseDC  (hdc);
}

// ---------------------------------------------------------

// ****************************************************
// Sets GpiPlayMetaFile options according to a mode key
// ****************************************************

void MetaDocument::setOpts(PLONG opts, MetaDocument::MetaMode mode) const
{
    switch(mode)
    {
        case MetaDocument::SETOPTS_PAINT:
           opts[PMF_SEGBASE]         = 0L;
           opts[PMF_LOADTYPE]        = LT_NOMODIFY;
           opts[PMF_RESOLVE]         = RS_DEFAULT;
           opts[PMF_LCIDS]           = LC_LOADDISC;
           opts[PMF_RESET]           = RES_DEFAULT;
           opts[PMF_SUPPRESS]        = SUP_DEFAULT;
           opts[PMF_COLORTABLES]     = CTAB_REPLACE;
           opts[PMF_COLORREALIZABLE] = CREA_DEFAULT;
           opts[PMF_DEFAULTS]        = DDEF_DEFAULT;
           break;

       case MetaDocument::SETOPTS_GETBOUNDS:
           opts[PMF_SEGBASE]         = 0L;
           opts[PMF_LOADTYPE]        = LT_NOMODIFY;
           opts[PMF_RESOLVE]         = RS_DEFAULT;
           opts[PMF_LCIDS]           = LC_LOADDISC;
           opts[PMF_RESET]           = RES_RESET;
           opts[PMF_SUPPRESS]        = SUP_SUPPRESS;
           opts[PMF_COLORTABLES]     = CTAB_REPLACE;
           opts[PMF_COLORREALIZABLE] = CREA_DEFAULT;
           opts[PMF_DEFAULTS]        = DDEF_DEFAULT;
           break;

       case MetaDocument::SETOPTS_GETSIZE:
       default:
           opts[PMF_SEGBASE]         = 0L;
           opts[PMF_LOADTYPE]        = LT_ORIGINALVIEW;
           opts[PMF_RESOLVE]         = RS_DEFAULT;
           opts[PMF_LCIDS]           = LC_LOADDISC;
           opts[PMF_RESET]           = RES_DEFAULT;
           opts[PMF_SUPPRESS]        = SUP_DEFAULT;
           opts[PMF_COLORTABLES]     = CTAB_REPLACE;
           opts[PMF_COLORREALIZABLE] = CREA_DEFAULT;
           opts[PMF_DEFAULTS]        = DDEF_LOADDISC;
           break;
    }
}


